/*
 *  V4L2 video capture example
 *
 *  This program can be used and distributed without restrictions.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>

#include <getopt.h>  /* getopt_long() */

#include <fcntl.h>  /* low-level i/o */
#include <unistd.h>
#include <errno.h>
#include <malloc.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <sys/time.h>
#include <sys/mman.h>
#include <sys/ioctl.h>

#include <errno.h>

#include <limits.h>

#include <asm/types.h>  /* for videodev2.h */

#include <linux/videodev2.h>

#include "png_io.h"

#define CLEAR(x) memset(&(x), 0, sizeof(x))

#define IMG 0
#define MIN_HOLD 1
#define PEAK_HOLD 2
#define AVE 3
#define DIF 4

#define FNAME_SIZE 160

struct v4l2_format fmt;

struct buffer {
  void *start;
  size_t length;
};

struct frame {
  unsigned char *img_buffer;
  int width;
  int height;
  time_t timestamp;
  long usec;
  long sum;
};

struct img_stack {
  unsigned long *img_buffer;
  int width;
  int height;
  time_t timestamp;
  long usec;
  int num_imgs;
  unsigned long min_val;
  unsigned long max_val;
};

/* Image frames */

/* Ring memory for triggered capture */
struct frame *frames = NULL;
/* Minimum stack */
struct frame min_hold;
/* Maximum stack */
struct frame peak_hold;
/* Average stack */
struct img_stack ave_stack;

static int img_in_mmap_buffer = 2;
static int img_in_buffer = 5;
static int n = 0;

/* Length of peak-hold/average/min -stack in seconds */
static int stack_length = 60;

static int previous_frame = -1;

static char *dev_name = NULL;
static int fd = -1;

/* MMAP buffers */
struct buffer *buffers = NULL;
static unsigned int n_buffers = 0;

static int threshold = 100000;

/* Which type of images will be saved */
static int use_min = 0;
static int use_max = 0;
static int use_ave = 0;
static int use_dif = 0;
static int use_trig = 0;

/* Time when the imaging started */
static time_t start_time;
/* Length of this imaging session */
static unsigned long capture_length = ULONG_MAX;
/* Starting time of the stacks */
static time_t stack_timestamp;

/* Output directory path. '/' will be added automatically */
static char *out_dir = NULL;

/* Prefix, which will be prepended to each filename. '_' will be added. */
static char *img_prefix = NULL;

static void errno_exit(const char *s){
  fprintf(stderr, "%s error %d, %s\n", s, errno, strerror(errno));
  exit(EXIT_FAILURE);
}

static int xioctl(int fd, int request, void *arg){
  int r;
  
  do {
    r = ioctl(fd, request, arg);
  } while(r == -1 && EINTR == errno);
  
  return r;
}

void make_fname(char *fname, time_t timestamp, long usec, int type){
  char tmp[50]; // = {'\0'};
  mode_t dir_mode = S_IRWXU |S_IRGRP | S_IXGRP | S_IROTH | S_IXOTH;

  memset(fname,'\0',FNAME_SIZE);

  if(out_dir != NULL){ strcat(fname,out_dir); strcat(fname,"/"); }

  strftime(tmp,50,"%Y",gmtime(&(timestamp))); strcat(fname,tmp); CLEAR(tmp);

  if(mkdir(fname, dir_mode) < 0 && errno != EEXIST){
    printf("Unable to create %s\nSaving to current direcory: %d\n",fname,errno);
    // Empty path
    memset(fname,'\0',FNAME_SIZE);
  } else {
    strftime(tmp,50,"/%m",gmtime(&(timestamp))); strcat(fname,tmp); CLEAR(tmp);
    mkdir(fname, dir_mode);
    strftime(tmp,50,"/%d",gmtime(&(timestamp))); strcat(fname,tmp); CLEAR(tmp);
    mkdir(fname, dir_mode);
    strcat(fname,"/");
  }

  if(img_prefix != NULL){ strcat(fname,img_prefix); strcat(fname,"_"); }

  strftime(tmp,50,"%Y-%m-%d_%H%M%S",gmtime(&(timestamp)));
  strcat(fname,tmp); CLEAR(tmp);

  switch(type){
  case AVE:
    sprintf(tmp,"_UTC_%d_s_ave.png", stack_length);
    break;
  case PEAK_HOLD:
    sprintf(tmp,"_UTC_%d_s_max.png", stack_length);
    break;
  case MIN_HOLD:
    sprintf(tmp,"_UTC_%d_s_min.png", stack_length);
    break;
  case DIF:
    sprintf(tmp,"_UTC_%d_s_dif.png", stack_length);
    break;
  default:
    sprintf(tmp,".%ld_UTC.png",usec);
    break;
  }

  strcat(fname,tmp);

}

void save_ave_stack(){

  int i;
  char *fname = (char*)malloc(FNAME_SIZE*sizeof(char));

  unsigned char *out_frame = calloc(ave_stack.width*ave_stack.height,sizeof(unsigned char));

  for(i=0;i<ave_stack.width*ave_stack.height;i++){
    //out_frame[i] = ave_stack.img_buffer[i]/ave_stack.num_imgs;
    //out_frame[i] = 255*ave_stack.img_buffer[i]/ave_stack.max_val;
    out_frame[i] = 255*(ave_stack.img_buffer[i]-ave_stack.min_val)/ave_stack.max_val;
  }
    
  make_fname(fname, ave_stack.timestamp, ave_stack.usec, AVE);
  printf("Saving %s, composed of %d frames\n",fname,ave_stack.num_imgs);
  writeImage(fname, ave_stack.width, ave_stack.height, out_frame, "This is my test image");
  
  free(out_frame);
  free(fname);

  free(ave_stack.img_buffer);
  ave_stack.img_buffer = NULL;
  ave_stack.min_val = ULONG_MAX;
  ave_stack.max_val = 0;
  ave_stack.num_imgs = 0;
}

void save_dif_stack(){

  int i;
  char *fname = (char*)malloc(FNAME_SIZE*sizeof(char));

  unsigned char *dif = calloc(min_hold.width*min_hold.height,sizeof(float));

  unsigned char *out_frame = calloc(min_hold.width*min_hold.height,sizeof(unsigned char));

  unsigned char dmin = UCHAR_MAX, dmax = 0;
  for(i=0;i<min_hold.width*min_hold.height;i++){
    dif[i] = peak_hold.img_buffer[i] - min_hold.img_buffer[i];
    if(dif[i] < dmin){ dmin = dif[i]; }
    if(dif[i] > dmax){ dmax = dif[i]; }
  }

  for(i=0;i<min_hold.width*min_hold.height;i++){
    out_frame[i] = 255*(dif[i]-dmin)/dmax;
  }
  
  free(dif); dif = NULL;

  make_fname(fname, min_hold.timestamp, min_hold.usec, DIF);
  printf("Saving %s\n",fname);
  writeImage(fname, min_hold.width, min_hold.height, out_frame, fname);
  
  free(out_frame); out_frame = NULL;
  free(fname); fname = NULL;

}

static void process_image(const void * p, time_t timestamp, long usec){

  static int packed_value;
  int i = 0, j = 0;
  int Y0, Y1, Y2, Y3;

  int x=0, y=0;
  int frame_ok = 0;
  unsigned long min = ULONG_MAX;

  if(use_trig){
    if(frames[n].img_buffer != NULL){
      free(frames[n].img_buffer);
      frames[n].img_buffer = NULL;
    }
    if(frames[n].img_buffer == NULL){
      frames[n].img_buffer = calloc(fmt.fmt.pix.width*fmt.fmt.pix.height,sizeof(*frames[n].img_buffer));
      frames[n].width = fmt.fmt.pix.width;
      frames[n].height = fmt.fmt.pix.height;
    }
    frames[n].sum = 0;  
  }

  if(use_max || use_dif){
    if(peak_hold.img_buffer == NULL){
      peak_hold.img_buffer = calloc(fmt.fmt.pix.width*fmt.fmt.pix.height,sizeof(*peak_hold.img_buffer));
      peak_hold.width = fmt.fmt.pix.width;
      peak_hold.height = fmt.fmt.pix.height;
      peak_hold.timestamp = timestamp;
      peak_hold.usec = usec;
      stack_timestamp = timestamp;
    }
  }

  if(use_min || use_dif){
    if(min_hold.img_buffer == NULL){
      min_hold.img_buffer = calloc(fmt.fmt.pix.width*fmt.fmt.pix.height,sizeof(*min_hold.img_buffer));
      for(i=0;i<fmt.fmt.pix.width*fmt.fmt.pix.height;i++){
	min_hold.img_buffer[i] = 255;
      }
      min_hold.width = fmt.fmt.pix.width;
      min_hold.height = fmt.fmt.pix.height;
      min_hold.timestamp = timestamp;
      min_hold.usec = usec;
      stack_timestamp = timestamp;
    }
  }

  if(use_ave){
    if(ave_stack.img_buffer == NULL){
      ave_stack.img_buffer = calloc(fmt.fmt.pix.width*fmt.fmt.pix.height, sizeof(*ave_stack.img_buffer));
      ave_stack.width = fmt.fmt.pix.width;
      ave_stack.height = fmt.fmt.pix.height;
      ave_stack.timestamp = timestamp;
      ave_stack.usec = usec;
      ave_stack.max_val = 0;
      ave_stack.min_val = ULONG_MAX;
      stack_timestamp = timestamp;
    }
  }

  i = 0;
  while(j+3 < (fmt.fmt.pix.width * fmt.fmt.pix.height)){
    packed_value = *((int*)p+i);

    switch(fmt.fmt.pix.pixelformat){
    case V4L2_PIX_FMT_GREY:
      Y0 = (char)(packed_value & 0xFF);
      Y2 = (char)((packed_value >> 4) & 0xFF);
      Y1 = (char)((packed_value >> 8) & 0xFF);
      Y3 = (char)((packed_value >> 12) & 0xFF);

      break;
    case V4L2_PIX_FMT_UYVY:
    case V4L2_PIX_FMT_YUYV:

      Y0 = (char)(packed_value & 0xFF);
      Y2 = (char)((packed_value >> 8) & 0xFF);
      Y1 = (char)((packed_value >> 16) & 0xFF);
      Y3 = (char)((packed_value >> 24) & 0xFF);

      break;
    default:
      break;
    }

    // Strip sign values after shift (i.e. unsigned shift)
    Y0 = Y0 & 0xFF;
    Y1 = Y1 & 0xFF;
    Y2 = Y2 & 0xFF;
    Y3 = Y3 & 0xFF;

    x = (i+1) % (fmt.fmt.pix.width/2) - 1;
    y = (i+1) / (fmt.fmt.pix.width/2);
    j = x*2 + y*fmt.fmt.pix.width;

    if(j+3 > fmt.fmt.pix.width * fmt.fmt.pix.height){ break; }

    if(use_trig){
      frames[n].img_buffer[j+0] = Y0;
      frames[n].img_buffer[j+1] = Y1;
      frames[n].img_buffer[j+2] = Y2;
      frames[n].img_buffer[j+3] = Y3;
    }

    if(use_max || use_dif){
      if(Y0 > peak_hold.img_buffer[j+0]){ peak_hold.img_buffer[j+0] = Y0; }
      if(Y1 > peak_hold.img_buffer[j+1]){ peak_hold.img_buffer[j+1] = Y1; }
      if(Y2 > peak_hold.img_buffer[j+2]){ peak_hold.img_buffer[j+2] = Y2; }
      if(Y3 > peak_hold.img_buffer[j+3]){ peak_hold.img_buffer[j+3] = Y3; }
    }

    if(use_min || use_dif){
      if(Y0 < min_hold.img_buffer[j+0]){ min_hold.img_buffer[j+0] = Y0; }
      if(Y1 < min_hold.img_buffer[j+1]){ min_hold.img_buffer[j+1] = Y1; }
      if(Y2 < min_hold.img_buffer[j+2]){ min_hold.img_buffer[j+2] = Y2; }
      if(Y3 < min_hold.img_buffer[j+3]){ min_hold.img_buffer[j+3] = Y3; }
    }

    if(use_ave){
      ave_stack.img_buffer[j+0] += Y0;
      ave_stack.img_buffer[j+1] += Y1;
      ave_stack.img_buffer[j+2] += Y2;
      ave_stack.img_buffer[j+3] += Y3;

      if(ave_stack.img_buffer[j+0] > ave_stack.max_val){
	ave_stack.max_val = ave_stack.img_buffer[j+0]; }
      if(ave_stack.img_buffer[j+1] > ave_stack.max_val){
	ave_stack.max_val = ave_stack.img_buffer[j+1]; }
      if(ave_stack.img_buffer[j+2] > ave_stack.max_val){
	ave_stack.max_val = ave_stack.img_buffer[j+2]; }
      if(ave_stack.img_buffer[j+3] > ave_stack.max_val){
	ave_stack.max_val = ave_stack.img_buffer[j+3]; }
      
      if(ave_stack.img_buffer[j+0] < min){ min = ave_stack.img_buffer[j+0]; }
      if(ave_stack.img_buffer[j+1] < min){ min = ave_stack.img_buffer[j+1]; }
      if(ave_stack.img_buffer[j+2] < min){ min = ave_stack.img_buffer[j+2]; }
      if(ave_stack.img_buffer[j+3] < min){ min = ave_stack.img_buffer[j+3]; }
    }

    if(Y0 > 0 && Y1 > 0 && Y2 > 0 && Y3 > 0){
      frame_ok = 1;
    }

    if(use_trig){
      frames[n].sum += Y0 + Y1 + Y2 + Y3;
    }

    i++;
  }

  if(previous_frame >= 0 && use_trig){
    if(frames[n].sum > frames[previous_frame].sum + threshold){
      printf("dif: %ld\n",frames[n].sum - frames[previous_frame].sum);
    }
  }

  previous_frame = n;
  // Increment index where to store image
  n++;
  // If we are using the last slot of the image buffer, start from the zero
  if(n >= img_in_buffer){
    n = 0;
  }

  if(frame_ok && use_ave){
    ave_stack.num_imgs++;

    if(min > ave_stack.min_val){
      ave_stack.min_val = min;
    }
  }

  if(stack_length < difftime(timestamp,stack_timestamp)){
    char *fname = NULL;

    if(use_ave){
      save_ave_stack();
    }

    if(use_dif){
      save_dif_stack();
    }

    if(use_max){
      fname = (char*)malloc(FNAME_SIZE*sizeof(char));
      make_fname(fname, peak_hold.timestamp, peak_hold.usec, PEAK_HOLD);
      printf("Saving %s\n",fname);
      writeImage(fname, peak_hold.width, peak_hold.height, peak_hold.img_buffer, "This is my test image");
      free(fname);
    }

    if(use_max || use_dif){
      free(peak_hold.img_buffer); peak_hold.img_buffer = NULL;
    }

    if(use_min){
      fname = (char*)malloc(FNAME_SIZE*sizeof(char));
      make_fname(fname, min_hold.timestamp, min_hold.usec, MIN_HOLD);
      printf("Saving %s\n",fname);
      writeImage(fname, min_hold.width, min_hold.height, min_hold.img_buffer, "This is my test image");
      free(fname);
      free(min_hold.img_buffer); min_hold.img_buffer = NULL;
    }

    if(use_min || use_dif){
      free(min_hold.img_buffer); min_hold.img_buffer = NULL;
    }
  }

}

static int read_frame(void){
  struct v4l2_buffer buf;

  time_t frametime;

  CLEAR(buf);
    
  buf.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
  buf.memory = V4L2_MEMORY_MMAP;
  
  if(xioctl(fd, VIDIOC_DQBUF, &buf) == -1) {
    switch(errno) {
    case EAGAIN:
      printf("EAGAIN\n");
      return 0;
      
    case EIO:
      /* Could ignore EIO, see spec. */
      printf("EIO\n");
      /* fall through */
      
    default:
      printf("default\n");
      errno_exit("VIDIOC_DQBUF");
    }
  }
    
  assert(buf.index < n_buffers);
  
  frametime = buf.timestamp.tv_sec;
  process_image(buffers[buf.index].start, frametime, buf.timestamp.tv_usec);
    
  if(xioctl(fd, VIDIOC_QBUF, &buf) == -1)
    errno_exit("VIDIOC_QBUF");
    

  return 1;
}

static void mainloop(void){
  time_t timenow;
  time(&start_time);
  time(&timenow);
  while(difftime(timenow,start_time) < capture_length) {
    for(;;) {
      fd_set fds;
      struct timeval tv;
      int r;
      
      FD_ZERO(&fds);
      FD_SET(fd, &fds);
      
      /* Timeout. */
      tv.tv_sec = 2;
      tv.tv_usec = 0;
      
      r = select(fd + 1, &fds, NULL, NULL, &tv);
      
      if(r == -1) {
	if(EINTR == errno)
	  continue;
	
	errno_exit("select");
      }
      
      if(r == 0) {
	fprintf(stderr, "select timeout\n");
	exit(EXIT_FAILURE);
      }
      
      if(read_frame())
	break;
      
      /* EAGAIN - continue select loop. */
    }
    time(&timenow);
  }
}

static void stop_capturing(void){
  enum v4l2_buf_type type;
  
  type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
  
  if(xioctl(fd, VIDIOC_STREAMOFF, &type) == -1)
    errno_exit("VIDIOC_STREAMOFF");

}

static void start_capturing(void){
  unsigned int i;
  enum v4l2_buf_type type;
  
  for(i = 0; i < n_buffers; ++i) {
    struct v4l2_buffer buf;
    
    CLEAR(buf);
    
    buf.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
    buf.memory = V4L2_MEMORY_MMAP;
    buf.index = i;
    
    if(xioctl(fd, VIDIOC_QBUF, &buf) == -1)
      errno_exit("VIDIOC_QBUF");
  }
  
  type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
  
  if(xioctl(fd, VIDIOC_STREAMON, &type) == -1)
    errno_exit("VIDIOC_STREAMON");
    
}

static void uninit_device(void){
  unsigned int i;
  
  for(i = 0; i < n_buffers; ++i)
    if(munmap(buffers[i].start, buffers[i].length) == -1)
      errno_exit("munmap");
  
  free(buffers);
}

static void init_mmap(int n){
  struct v4l2_requestbuffers req;
  
  CLEAR(req);
  
  req.count = 2;
  req.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
  req.memory = V4L2_MEMORY_MMAP;
  
  if(xioctl(fd, VIDIOC_REQBUFS, &req) == -1) {
    if(EINVAL == errno) {
      fprintf(stderr, "%s does not support "
	      "memory mapping\n", dev_name);
      exit(EXIT_FAILURE);
    } else {
      errno_exit("VIDIOC_REQBUFS");
    }
  }
  
  if(req.count < 2) {
    fprintf(stderr, "Insufficient buffer memory on %s\n",
	    dev_name);
    exit(EXIT_FAILURE);
  }
  
  buffers = calloc(req.count, sizeof(*buffers));
  
  if(!buffers) {
    fprintf(stderr, "Out of memory\n");
    exit(EXIT_FAILURE);
  }
  
  for(n_buffers = 0; n_buffers < req.count; ++n_buffers) {
    struct v4l2_buffer buf;
    
    CLEAR(buf);
    
    buf.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
    buf.memory = V4L2_MEMORY_MMAP;
    buf.index = n_buffers;
    
    if(xioctl(fd, VIDIOC_QUERYBUF, &buf) == -1)
      errno_exit("VIDIOC_QUERYBUF");
    
    buffers[n_buffers].length = buf.length;
    buffers[n_buffers].start =
      mmap(NULL /* start anywhere */,
	   buf.length,
	   PROT_READ | PROT_WRITE /* required */,
	   MAP_SHARED /* recommended */,
	   fd, buf.m.offset);
    
    if(MAP_FAILED == buffers[n_buffers].start)
      errno_exit("mmap");
  }
}

static void init_device(int img_in_mmap_buffer){
  struct v4l2_capability cap;
  struct v4l2_cropcap cropcap;
  struct v4l2_crop crop;
  unsigned int min;
  
  if(xioctl(fd, VIDIOC_QUERYCAP, &cap) == -1) {
    if(EINVAL == errno) {
      fprintf(stderr, "%s is no V4L2 device\n", dev_name);
      exit(EXIT_FAILURE);
    } else {
      errno_exit("VIDIOC_QUERYCAP");
    }
  }
  
  if(!(cap.capabilities & V4L2_CAP_VIDEO_CAPTURE)) {
    fprintf(stderr, "%s is no video capture device\n",
	    dev_name);
    exit(EXIT_FAILURE);
  }
  
  if(!(cap.capabilities & V4L2_CAP_STREAMING)) {
    fprintf(stderr, "%s does not support streaming i/o\n",
	    dev_name);
    exit(EXIT_FAILURE);
  }
  
  
  /* Select video input, video standard and tune here. */
  
  CLEAR(cropcap);

  cropcap.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
  
  if(xioctl(fd, VIDIOC_CROPCAP, &cropcap) == 0) {
    crop.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
    crop.c = cropcap.defrect; /* reset to default */
    
    if(xioctl(fd, VIDIOC_S_CROP, &crop) == -1) {
      switch(errno) {
      case EINVAL:
	/* Cropping not supported. */
	break;
      default:
	/* Errors ignored. */
	break;
      }
    }
  } else {
    /* Errors ignored. */
  }
  
  
  CLEAR(fmt);
  
  fmt.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
  fmt.fmt.pix.width = 720;
  fmt.fmt.pix.height = 576;
  fmt.fmt.pix.pixelformat = V4L2_PIX_FMT_YUYV;
  fmt.fmt.pix.field = V4L2_FIELD_INTERLACED;

  if(xioctl(fd, VIDIOC_S_FMT, &fmt) == -1)
    errno_exit("VIDIOC_S_FMT");
  
  /* Note VIDIOC_S_FMT may change width and height. */
  
  /* Buggy driver paranoia. */
  min = fmt.fmt.pix.width * 2;
  if(fmt.fmt.pix.bytesperline < min)
    fmt.fmt.pix.bytesperline = min;
  min = fmt.fmt.pix.bytesperline * fmt.fmt.pix.height;
  if(fmt.fmt.pix.sizeimage < min)
    fmt.fmt.pix.sizeimage = min;
  
  init_mmap(img_in_mmap_buffer);

}

static void close_device(void){
  if(close(fd) == -1)
    errno_exit("close");
  
  fd = -1;
}

static void open_device(void){
  struct stat st; 
  
  if(stat(dev_name, &st) == -1) {
    fprintf(stderr, "Cannot identify '%s': %d, %s\n",
	    dev_name, errno, strerror(errno));
    exit(EXIT_FAILURE);
  }
  
  if(!S_ISCHR(st.st_mode)) {
    fprintf(stderr, "%s is no device\n", dev_name);
    exit(EXIT_FAILURE);
  }
  
  fd = open(dev_name, O_RDWR /* required */ | O_NONBLOCK, 0);
  
  if(fd == -1) {
    fprintf(stderr, "Cannot open '%s': %d, %s\n",
	    dev_name, errno, strerror(errno));
    exit(EXIT_FAILURE);
  }
}

static void usage (FILE *fp, int argc, char **argv){
  fprintf(fp,
	  "Usage: %s [options]\n\n"
	  "Options:\n"
	  "-c | --capture-device <name> Video device name [/dev/video0]\n"
	  "-h | --help  Print this message\n"
	  "-n | --buffer <n>  Number of images in ring buffer [10]\n"
	  "                   Enables motion detection\n"
	  "-s | --stack-length <s>  Length of stacked period in seconds [60]\n"
	  "-m | --min  Enable minimum stacking [default: off]\n"
	  "-M | --max  Enable maximum stacking [default: off]\n"
	  "-a | --ave  Enable average stacking [default: off]\n"
	  "-d | --dif  Enable difference stacking [default: off]\n"
	  "-t | --trig  Enable motion detection with default ring buffer [default: off]\n"
	  "-l | --capture-length <time>  Length of imaging period in seconds [unlimited]\n"
	  "-p | --prefix <prefix>  Image filename prefix [default: none]\n"
	  "-o | --out-dir <dir>  Image output directory [default: current directory]\n"
	  "                      The images are saved to <dir>/yyyy/mm/dd/\n"
	  "\n",
	  argv[0]);
}

static const char short_options [] = "c:hn:s:mMadtp:o:l:";

static const struct option long_options [] = {
  { "capture-device", required_argument, NULL, 'c' }, 
  { "help", no_argument, NULL, 'h' }, 
  { "buffer", required_argument, NULL, 'n' }, 
  { "stack-length", required_argument, NULL, 's' }, 
  { "min", no_argument, NULL, 'm' }, 
  { "max", no_argument, NULL, 'M' }, 
  { "ave", no_argument, NULL, 'a' }, 
  { "dif", no_argument, NULL, 'd' }, 
  { "trig", no_argument, NULL, 't' },
  { "prefix", required_argument, NULL, 'p' }, 
  { "out-dir", required_argument, NULL, 'o' }, 
  { "capture-length", required_argument, NULL, 'l' }, 
  { 0, 0, 0, 0 }
};

int main(int argc, char **argv){
  dev_name = "/dev/video0";

  for(;;) {
    int index;
    int c;
    
    c = getopt_long(argc, argv,
		    short_options, long_options,
		    &index);
    
    if(c == -1)
      break;
    
    switch(c) {
    case 0: /* getopt_long() flag */
      break;
    case 'c':
      dev_name = optarg;
      break;      
    case 'h':
      usage(stdout, argc, argv);
      exit(EXIT_SUCCESS);
    case 'n':
      sscanf(optarg,"%d",&img_in_buffer);
      use_trig = 1;
      break;
    case 's':
      sscanf(optarg,"%d",&stack_length);
      break;
    case 'm':
      use_min = 1;
      break;
    case 'M':
      use_max = 1;
      break;
    case 'a':
      use_ave = 1;
      break;
    case 't':
      use_trig = 1;
      break;
    case 'd':
      use_dif = 1;
      break;
    case 'l':
      sscanf(optarg,"%ld",&capture_length);
      break;
    case 'p':
      img_prefix = optarg;
      break;
    case 'o':
      out_dir = optarg;
      break;      
    default:
      usage(stderr, argc, argv);
      exit(EXIT_FAILURE);
    }
  }

  if(!use_min && !use_max && !use_ave && !use_trig && !use_dif){
    printf("\nNothing to do, terminating.\n");
    printf("Specify at least one imaging mode!\n\n");
    usage(stderr, argc, argv);
    exit(EXIT_FAILURE);
  }

  int i;
  ave_stack.img_buffer = NULL;
  ave_stack.max_val = 0;
  ave_stack.max_val = ULONG_MAX;
  
  peak_hold.img_buffer = NULL;
  min_hold.img_buffer = NULL;

  if(use_trig){
    frames = calloc(img_in_buffer,sizeof(*frames));
    for(i=0; i<img_in_buffer; i++){
      frames[i].img_buffer = NULL;
      frames[i].width = -1;
      frames[i].height = -1;
    }
  }

  open_device();

  init_device(img_in_mmap_buffer);

  start_capturing();

  mainloop();

  stop_capturing();

  uninit_device();

  close_device();

  if(use_trig){
    for(i=0; i < img_in_buffer; i++){
      if(frames[i].img_buffer != NULL){
	free(frames[i].img_buffer);
	frames[i].img_buffer = NULL;
      }
    }
  }

  if(frames != NULL){ free(frames); }
  if(peak_hold.img_buffer != NULL){ free(peak_hold.img_buffer); }
  if(min_hold.img_buffer != NULL){ free(min_hold.img_buffer); }
  if(ave_stack.img_buffer != NULL){ free(ave_stack.img_buffer); }

  exit(EXIT_SUCCESS);

  return 0;
}
